%{
#define ZASM_LEXER_C 1
#include <zasm/grow_string.h>
#include <zasm/errors.h>
#include <zasm/memory.h>
#include <zasm/string.h>
#include <zasm/lexer.h>
#include <stdio.h>

#define YY_DECL struct token_t lexer_internal(void)
%}

%x STRING
%x ISTRING
%x INCLUDE

DIR [ \t]*[.#][ \t]*
SPACES [ \t]*
%%
%{
  static struct grow_string_t str = GROW_STRING_EMPTY;
  static unsigned int num;
  struct token_t token;
  grow_string_reset(&str);
%}
  /* auto-gen lexer */
"LITTLE ENDIAN" {token.type = TOK_LITTLE_ENDIAN;{}return token;}
"BIG ENDIAN" {token.type = TOK_BIG_ENDIAN;{}return token;}
\`[0-9a-fA-F]{2}\` {token.type = TOK_BYTE_INT;{token.value.t_int = strtol(yytext+1,NULL,16);}return token;}
\`[0-9a-fA-F]{4}\` {token.type = TOK_WORD_INT;{token.value.t_int = strtol(yytext+1,NULL,16);}return token;}
\`[0-9a-fA-F]{8}\` {token.type = TOK_DWORD_INT;{token.value.t_int = strtol(yytext+1,NULL,16);}return token;}
{DIR}byte {token.type = TOK_BYTE;{}return token;}
{DIR}word {token.type = TOK_WORD;{}return token;}
{DIR}dword {token.type = TOK_DWORD;{}return token;}
@[0-9]+ {token.type = TOK_ARG;{token.value.t_int = strtol(yytext+1, NULL,16);}return token;}
{DIR}[dD][bB] {token.type = TOK_DB;{}return token;}
[dD][bB] {token.type = TOK_DB;{}return token;}
{DIR}[dD][wW] {token.type = TOK_DW;{}return token;}
[dD][wW] {token.type = TOK_DW;{}return token;}
{DIR}[dD][dD] {token.type = TOK_DD;{}return token;}
[dD][dD] {token.type = TOK_DD;{}return token;}
<<EOF>> {token.type = TOK_END_OF_FILE;{}return token;}
[a-zA-Z0-9_]+[ \t]*\.\.\. {token.type = TOK_VARARG;{for (num = 0; ((('0'<=yytext[num]) && (yytext[num]<='9'))||
                  (('a'<=yytext[num]) && (yytext[num]<='z'))||
                  (('A'<=yytext[num]) && (yytext[num]<='Z'))||
                  (yytext[num] == '_')); num++);
   yytext[num] = 0;
   token.value.t_char_p = string_new(yytext);}return token;}
(;[^\n]*)?[ \t]*\r?\n {token.type = TOK_NEWLINE;{}return token;}
\\\r?\n {token.type = TOK_ESCAPED_NEWLINE;{}return token;}
[ \t]+ {token.type = TOK_SPACE;{}return token;}
"%"[01]+ {token.type = TOK_INTEGER;{token.value.t_int = strtol(yytext+1, NULL, 2);}return token;}
0[bB][01]+ {token.type = TOK_INTEGER;{token.value.t_int = strtol(yytext+2, NULL, 2);}return token;}
[01]+[bB] {token.type = TOK_INTEGER;{token.value.t_int = strtol(yytext, NULL, 2);}return token;}
"0"[0-7]+ {token.type = TOK_INTEGER;{token.value.t_int = strtol(yytext+1, NULL, 8);}return token;}
0[oO][0-7]+ {token.type = TOK_INTEGER;{token.value.t_int = strtol(yytext+2, NULL, 8);}return token;}
[0-7]+[oO] {token.type = TOK_INTEGER;{token.value.t_int = strtol(yytext, NULL, 8);}return token;}
[0-9]+ {token.type = TOK_INTEGER;{token.value.t_int = strtol(yytext, NULL, 10);}return token;}
0[dD][0-7]+ {token.type = TOK_INTEGER;{token.value.t_int = strtol(yytext+2, NULL, 10);}return token;}
[0-7]+[dD] {token.type = TOK_INTEGER;{token.value.t_int = strtol(yytext, NULL, 10);}return token;}
"$"[0-9a-fA-F]+ {token.type = TOK_INTEGER;{token.value.t_int = strtol(yytext+1, NULL, 16);}return token;}
0[xXhH][0-9a-fA-F]+ {token.type = TOK_INTEGER;{token.value.t_int = strtol(yytext+2, NULL, 16);}return token;}
[0-9a-fA-F]+[hH] {token.type = TOK_INTEGER;{token.value.t_int = strtol(yytext, NULL, 16);}return token;}
{DIR}align_addr {token.type = TOK_ALIGN_ADDR;{}return token;}
{DIR}align_mult {token.type = TOK_ALIGN_MULT;{}return token;}
{DIR}org {token.type = TOK_ORG;{}return token;}
{DIR}setpage {token.type = TOK_PAGE;{}return token;}
{DIR}page_start {token.type = TOK_PAGE_START;{}return token;}
{DIR}newpage {token.type = TOK_NEWPAGE;{}return token;}
{DIR}(equ|EQU) {token.type = TOK_EQU;{}return token;}
equ|EQU {token.type = TOK_EQU;{}return token;}
{DIR}export {token.type = TOK_EXPORT;{}return token;}
{DIR}begin_section {token.type = TOK_BEGIN_SECTION;{}return token;}
{DIR}end_section {token.type = TOK_END_SECTION;{}return token;}
\\ {token.type = TOK_BACKSLASH;{}return token;}
\? {token.type = TOK_OP_IFTHEN;{}return token;}
: {token.type = TOK_OP_ELSE;{}return token;}
"||" {token.type = TOK_OP_LOG_OR;{}return token;}
"&&" {token.type = TOK_OP_LOG_AND;{}return token;}
"^^" {token.type = TOK_OP_LOG_XOR;{}return token;}
"!!"|"!"|"~~" {token.type = TOK_OP_LOG_NOT;{}return token;}
"|"|".|" {token.type = TOK_OP_BIN_OR;{}return token;}
".&" {token.type = TOK_OP_BIN_AND;{}return token;}
"^"|".^" {token.type = TOK_OP_BIN_XOR;{}return token;}
".!"|".~"|"~" {token.type = TOK_OP_BIN_NOT;{}return token;}
"<<"|".<" {token.type = TOK_OP_SHIFT_LEFT;{}return token;}
">>"|".>" {token.type = TOK_OP_SHIFT_RIGHT;{}return token;}
"=="|"=" {token.type = TOK_OP_EQU;{}return token;}
"!="|"/="|"<>" {token.type = TOK_OP_NEQ;{}return token;}
"<" {token.type = TOK_OP_LT;{}return token;}
">" {token.type = TOK_OP_GT;{}return token;}
"<="|"=<" {token.type = TOK_OP_LE;{}return token;}
">="|"=>" {token.type = TOK_OP_GE;{}return token;}
"<=>"|">=<" {token.type = TOK_OP_CMP;{}return token;}
"+" {token.type = TOK_OP_ADD;{}return token;}
"-" {token.type = TOK_OP_SUB;{}return token;}
"*" {token.type = TOK_OP_MUL;{}return token;}
"/" {token.type = TOK_OP_DIV;{}return token;}
"%" {token.type = TOK_OP_MOD;{}return token;}
"**" {token.type = TOK_OP_POW;{}return token;}
"." {token.type = TOK_DOT;{}return token;}
\( {token.type = TOK_PAREN_OPEN;{}return token;}
\) {token.type = TOK_PAREN_CLOSE;{}return token;}
\{ {token.type = TOK_BRACE_OPEN;{}return token;}
\} {token.type = TOK_BRACE_CLOSE;{}return token;}
\[ {token.type = TOK_SQUARE_OPEN;{}return token;}
\] {token.type = TOK_SQUARE_CLOSE;{}return token;}
, {token.type = TOK_COMMA;{}return token;}
"##"|"&" {token.type = TOK_PASTE;{}return token;}
defined {token.type = TOK_DEFINED;{}return token;}
{DIR}(define|def|declare) {token.type = TOK_DEFINE;{}return token;}
{DIR}(macro) {token.type = TOK_MACRO;{}return token;}
macro {token.type = TOK_MACRO;{}return token;}
{DIR}(redefine|redef|redeclare) {token.type = TOK_REDEF;{}return token;}
{DIR}(undefine|undef|undeclare) {token.type = TOK_UNDEF;{}return token;}
{DIR}(endm|endmacro|enddefine|enddef|enddeclare) {token.type = TOK_ENDM;{}return token;}
{DIR}(pp_begin_section) {token.type = TOK_PP_BEGIN_SECTION;{}return token;}
{DIR}(pp_end_section) {token.type = TOK_PP_END_SECTION;{}return token;}
{DIR}(include|input) {token.type = TOK_INCLUDE;{BEGIN(INCLUDE);}return token;}
{DIR}(if) {token.type = TOK_IF;{}return token;}
{DIR}(then) {token.type = TOK_THEN;{}return token;}
{DIR}(ifdef) {token.type = TOK_IFDEF;{}return token;}
{DIR}(ifndef) {token.type = TOK_IFNDEF;{}return token;}
{DIR}(else) {token.type = TOK_ELSE;{}return token;}
{DIR}(elseif|elsif) {token.type = TOK_ELSEIF;{}return token;}
{DIR}(elseifdef|elsifdef) {token.type = TOK_ELSEIFDEF;{}return token;}
{DIR}(elseifndef|elsifndef) {token.type = TOK_ELSEIFNDEF;{}return token;}
{DIR}(endif) {token.type = TOK_ENDIF;{}return token;}
{DIR}(for) {token.type = TOK_FOR;{}return token;}
{DIR}(in) {token.type = TOK_IN;{}return token;}
{DIR}(from) {token.type = TOK_FROM;{}return token;}
{DIR}(to) {token.type = TOK_TO;{}return token;}
{DIR}(step) {token.type = TOK_STEP;{}return token;}
{DIR}(do) {token.type = TOK_DO;{}return token;}
{DIR}(endfor) {token.type = TOK_ENDFOR;{}return token;}
{DIR}(error) {token.type = TOK_ERROR;{}return token;}
{DIR}(warning) {token.type = TOK_WARNING;{}return token;}
{DIR}(debug) {token.type = TOK_DEBUG;{}return token;}
{DIR}output {token.type = TOK_OUTPUT;{}return token;}


  /* strings */
<INCLUDE>{SPACES} {token.type = TOK_SPACE; return token;}
<INCLUDE>"<"         BEGIN ISTRING;
<ISTRING>">"         {
  BEGIN INITIAL;
  token.type = TOK_ISTRING;
  if (str.string)
    token.value.t_char_p = string_new(str.string);
  else
    token.value.t_char_p = string_new("");
  return token;
  }
<INITIAL,INCLUDE>\"  BEGIN STRING;
<STRING>\"           {
  BEGIN(INITIAL);
  token.type = TOK_STRING;
  if (str.string)
    token.value.t_char_p = string_new(str.string);
  else
    token.value.t_char_p = string_new("");
  return token;
  }
<STRING>[^\\\"\n]+            grow_string_add(&str, yytext);
<ISTRING>[^\\\>\n]+           grow_string_add(&str, yytext);
<ISTRING,STRING>\n            ERROR("unterminated string\n");
<ISTRING,STRING>\\\n          ERROR("unterminated string\n");
<ISTRING,STRING>\\n           grow_string_add_char(&str, '\n');
<ISTRING,STRING>\\t           grow_string_add_char(&str, '\t');
<ISTRING,STRING>\\r           grow_string_add_char(&str, '\r');
<ISTRING,STRING>\\e           grow_string_add_char(&str, '\e');
<ISTRING,STRING>\\f           grow_string_add_char(&str, '\f');
<ISTRING,STRING>\\a           grow_string_add_char(&str, '\a');
<ISTRING,STRING>\\[0-7]{1,3}  {sscanf(yytext+1, "%o", &num); grow_string_add_char(&str, num);}
<ISTRING,STRING>\\[oO][0-7]{1,3}  {sscanf(yytext+2, "%o", &num); grow_string_add_char(&str, num);}
<ISTRING,STRING>\\[dD][0-9]{1,3}  {sscanf(yytext+2, "%u", &num); grow_string_add_char(&str, num);}
<ISTRING,STRING>\\[xXhH][0-7]{1,2}  {sscanf(yytext+2, "%x", &num); grow_string_add_char(&str, num);}
<ISTRING,STRING>\\.           grow_string_add_char(&str, yytext[1]);

  /* names */
[0-9a-zA-Z_\']+             {token.type = TOK_NAME; token.value.t_char_p = string_new(yytext); return token;}
[\$\@]             {token.type = TOK_NAME; token.value.t_char_p = string_new(yytext); return token;}
  /* special chars */
.                     {token.type = TOK_SPECIAL; token.value.t_char = yytext[0]; return token;}

%%
struct lexer_t *lexer_new(char *file)
{
  struct lexer_t *val;
  ALLOCATE(struct lexer_t, val, 1);
  lexer_fill(val, file);
  return val;
}

void lexer_fill(struct lexer_t *lexer, char *filename)
{
  FILE *file;
  file = fopen(filename, "r");
  if (!file)
    ERROR("can not open file \"%s\"\n", filename);
  lexer->file = file;
  lexer->position.file = string_copy(filename);
  lexer->position.line_begin = 1;
  lexer->position.line_end = 1;
  lexer->buffer = yy_create_buffer(file, YY_BUF_SIZE);
  lexer->start_condition = INITIAL;
}

void lexer_execute(struct feeder_t *in, struct stream_t *out, struct lexer_t *lexer)
{
  int i;
  struct token_t token;
  struct position_t pos;

  pos = lexer->position;
  yy_switch_to_buffer(lexer->buffer);
  BEGIN(lexer->start_condition);
  for (i = 0; i < 20; i++)
    {
      token = lexer_internal();
      pos.line_begin = pos.line_end;
      if (token.type == TOK_NEWLINE)
	pos.line_end++;
      if (token.type == TOK_ESCAPED_NEWLINE)
	{
	  pos.line_end++;
	  continue;
	}
      token.position = pos;
      stream_put(out, token);
      if (token.type == TOK_END_OF_FILE)
	{
	  lexer_free(lexer);
	  return;
	}
    }
  
  lexer->start_condition = YY_START;
  lexer->buffer = YY_CURRENT_BUFFER;
  lexer->position = pos;
}

void lexer_delete(struct lexer_t *lexer)
{
  if (fclose(lexer->file))
    ERROR("I/O error\n");
  string_free(lexer->position.file);
  yy_delete_buffer(lexer->buffer);
}

void lexer_free(struct lexer_t *lexer)
{
  lexer_delete(lexer);
  free(lexer);
}

struct string_lexer_t *string_lexer_new(char *input)
{
  struct string_lexer_t *val;
  ALLOCATE(struct string_lexer_t, val, 1);
  string_lexer_fill(val, input);
  return val;
}

void string_lexer_fill(struct string_lexer_t *lexer, char *input)
{
  lexer->position.file = string_new("no file");
  lexer->position.line_begin = 1;
  lexer->position.line_end = 1;
  lexer->buffer = yy_scan_string(input);
  lexer->start_condition = INITIAL;
}

void string_lexer_execute(struct feeder_t *in, struct stream_t *out, struct string_lexer_t *lexer)
{
  int i;
  struct token_t token;
  struct position_t pos;

  pos = lexer->position;
  yy_switch_to_buffer(lexer->buffer);
  BEGIN(lexer->start_condition);
  for (i = 0; i < 20; i++)
    {
      token = lexer_internal();
      pos.line_begin = pos.line_end;
      if (token.type == TOK_NEWLINE)
	pos.line_end++;
      if (token.type == TOK_ESCAPED_NEWLINE)
	{
	  pos.line_end++;
	  continue;
	}
      token.position = pos;
      stream_put(out, token);
      if (token.type == TOK_END_OF_FILE)
	{
	  string_lexer_free(lexer);
	  return;
	}
    }
  
  lexer->start_condition = YY_START;
  lexer->buffer = YY_CURRENT_BUFFER;
  lexer->position = pos;
}

void string_lexer_delete(struct string_lexer_t *lexer)
{
  string_free(lexer->position.file);
  yy_delete_buffer(lexer->buffer);
}

void string_lexer_free(struct string_lexer_t *lexer)
{
  string_lexer_delete(lexer);
  free(lexer);
}
