#define ZASM_LINKER_C 1

#include <zasm/linker.h>
#include <zasm/memory.h>

struct linker_t *linker_new(char *dump_file, char *export_file)
{
  struct linker_t *val;
  ALLOCATE(struct linker_t, val, 1);
  linker_fill(val, dump_file, export_file);
  return val;
}

void linker_fill(struct linker_t *lnk, char *dump_file, char *export_file)
{
  lnk->address = 0;
  lnk->page = 0;
  lnk->page_start = 0x4000;
  lnk->look_ahead.type = TOKENS_COUNT;
  lnk->expr_alloc = NULL;
  lnk->int_alloc = int_alloc_new();
  lnk->ns = link_namespace_new();
  link_namespace_add_level(lnk->ns);
  lnk->dump_file = dump_file;
  if (dump_file)
    lnk->dump_labels = list_new();
  else
    lnk->dump_labels = NULL;
  lnk->export_file = export_file;
  if (export_file)
    lnk->export_labels = list_new();
  else
    lnk->export_labels = NULL;
}

int *linker_get_int(struct linker_t *lnk)
{
  return int_alloc_get(&(lnk->int_alloc));
}
// linker_execute is in lnk.parser

void linker_delete(struct linker_t *lnk)
{
  link_namespace_del_level(lnk->ns);
  if (!link_namespace_is_empty(lnk->ns))
    ERROR("reached end of input before #begin_section was complete\n");
  link_namespace_free(lnk->ns);
  expr_alloc_free(lnk->expr_alloc);
  if (lnk->dump_labels)
    list_free(lnk->dump_labels);
  if (lnk->export_labels)
    list_free(lnk->export_labels);
}

void linker_free(struct linker_t *lnk)
{
  linker_delete(lnk);
  free(lnk);
}

struct int_alloc_t *int_alloc_new()
{
  struct int_alloc_t *val;
  ALLOCATE(struct int_alloc_t, val, 1);
  int_alloc_fill(val);
  return val;
}

void int_alloc_fill(struct int_alloc_t *a)
{
  a->previous = NULL;
  a->i = 0;
}

int *int_alloc_get(struct int_alloc_t **a)
{
  struct int_alloc_t *t;
  if ((*a)->i >= 1024)
    {
      t = int_alloc_new();
      t->previous = *a;
      *a = t;
    }
  (*a)->i++;
  return &((*a)->data[(*a)->i-1]);
}

void int_alloc_reset(struct int_alloc_t *a)
{
  int_alloc_delete(a);
  int_alloc_fill(a);
}

void int_alloc_delete(struct int_alloc_t *a)
{
  if (a->previous)
    int_alloc_free(a->previous);
}

void int_alloc_free(struct int_alloc_t *a)
{
  int_alloc_delete(a);
  free(a);
}
