#define ZASM_LINK_ENTRY_C 1

#include <zasm/link_expr.h>
#include <zasm/link_entry.h>
#include <zasm/string_alloc.h>

ALLOC_DEFINE(link_entry, ALLOC_DEF_COUNT);

struct link_entry_t *link_entry_new(struct link_level_t *level, char *name)
{
  struct link_entry_t *val;

  val = link_entry_alloc_get(&(level->entry_alloc));
  link_entry_fill(val, level, name);
  return val;
}

void link_entry_fill(struct link_entry_t *entry, struct link_level_t *level, char *name)
{
  char *cname;

  cname = string_alloc_copy(&(level->string_alloc), name);
  hash_entry_fill(&(entry->entry), cname);
  entry->page = 0;
  entry->flags = 0;
  entry->expr = NULL;
}

struct link_entry_t *link_entry_copy(struct link_entry_t *orig, struct link_level_t *to)
{
  return link_entry_new(to, orig->entry.name);
}

void link_entry_define(struct link_entry_t *entry, struct link_level_t *level, struct expr_t *expr, int addr, int page)
{
  int flags = entry->flags;
  entry->flags = LINK_ENTRY_DEFINED; // also clear global
  entry->page = page;
  entry->expr = link_expr_from_expr(level, expr, addr, page);
  if (flags&LINK_ENTRY_GLOBAL)
    link_namespace_do_export(LINK_NAMESPACE_T(level->level.parent), entry);
}

void link_entry_define_global(struct link_entry_t *entry, int page)
{
  entry->flags = LINK_ENTRY_DEFINED; // clear global
  entry->page = page;
}

void link_entry_set_global(struct link_entry_t *entry)
{
  entry->flags |= LINK_ENTRY_GLOBAL;
}

int link_entry_defined(struct link_entry_t *entry)
{
  return (entry->flags) & LINK_ENTRY_DEFINED;
}

int link_entry_global(struct link_entry_t *entry)
{
  return (entry->flags) & LINK_ENTRY_GLOBAL;
}

void link_entry_delete(struct link_entry_t *entry)
{
  hash_entry_delete(&(entry->entry));
}

void link_entry_free(struct link_entry_t *entry)
{
  link_entry_delete(entry);
}
