#define ZASM_PARSER_C 1

#include <zasm/tokens.h>
#include <zasm/parser.h>
#include <zasm/list.h>
#include <zasm/string.h>
#include <zasm/str.h>
#include <string.h>
#include <stdarg.h>

static void paste_tokens(struct list_t *dest, struct token_t t1, struct token_t t2)
{
  char *val;

  if ((t1.type == TOK_NAME)&&((t2.type == TOK_NAME)||(t2.type == TOK_VARARG)))
    {
      val = STR_CONCAT(t1.value.t_char_p, t2.value.t_char_p);
      t2.value.t_char_p = string_new(val);
      list_add_end(dest, t2);
    }
  else if ((t1.type == TOK_NAME)&&(t2.type == TOK_INTEGER))
    {
      val = str_append_int(t1.value.t_char_p, t2.value.t_int);
      t1.value.t_char_p = string_new(val);
      list_add_end(dest, t1);
    }
  else if ((t1.type == TOK_INTEGER)&&(t2.type == TOK_NAME))
    {
      val = str_prepend_int(t1.value.t_int, t2.value.t_char_p);
      t1.value.t_char_p = string_new(val);
      list_add_end(dest, t1);
    }
  else if ((t1.type == TOK_COMMA) && (t2.type == TOK_COMMA))
    list_add_end(dest, t1);
  else if ((t1.type == TOK_COMMA) && (t2.type == TOK_PAREN_CLOSE))
    list_add_end(dest, t2);
  else if ((t1.type == TOK_PAREN_OPEN) && (t2.type == TOK_COMMA))
    list_add_end(dest, t1);
  else
    {
      if (t1.type != TOK_LIST_END)
	list_add_end(dest,token_copy(t1));
      if (t2.type != TOK_LIST_END)
	list_add_end(dest,token_copy(t2));
    }
}

#define SUBSTITUTE(SUBS,T1) \
  foreach (token,body)\
    {\
      if (token.type == TOK_PASTE)\
	{\
	  T1 = list_get_end(value);\
	  while (T1.type == TOK_SPACE)\
	    T1 = list_get_end(value);\
	  token = list_read_begin(body);\
	  while (token.type == TOK_SPACE)\
	    token = list_read_begin(body);\
	  paste = 1;\
	}\
      if (token.type == TOK_NAME)\
	{\
          SUBS;\
	}\
      if (paste)\
	{\
	  while (token.type == TOK_SPACE)\
	    token = list_read_begin(body);\
	  paste_tokens(value,T1, token);\
	  paste = 0;\
	}\
      else\
        list_add_end(value, token_copy(token));\
    }

struct list_t *substitute_macro(struct list_t *body, struct list_t *arg_names, struct list_t *arg_values)
{
  static struct pp_arg_t *args = NULL;
  int args_count = 0;
  static int args_mem = 0;

  struct list_t *value;
  struct list_t *temp;
  struct token_t arg_name;
  struct token_t arg_value;
  struct token_t token;
  int i;
  int paste = 0;
  
  token.type = TOK_COMMA;
  token.position.file = "";
  token.position.line_begin = 0;
  token.position.line_end = 0;

  foreach (arg_name,arg_names)
    {
      while (args_count >= args_mem)
	{
	  args_mem+=10;
	  REALLOCATE(struct pp_arg_t, args, args_mem);
	}
      args[args_count].name = arg_name.value.t_char_p;
      arg_value = list_get_begin(arg_values);
      if (arg_value.type == TOK_LIST_END)
	if (arg_name.type == TOK_VARARG)
	  args[args_count].value = list_new();
	else
	  T_ERROR(arg_name, "not enough arguments for a macro, or vararg is not last\n");
      else
	args[args_count].value = arg_value.value.t_struct_list_t_p;
      if (arg_name.type == TOK_VARARG)
	for (arg_value = list_get_begin(arg_values); arg_value.type != TOK_LIST_END; arg_value = list_get_begin(arg_values))
	  {
	    list_add_end(args[args_count].value, token);
	    list_add_list_end(args[args_count].value, arg_value.value.t_struct_list_t_p);
	  }
      args_count++;
    }
  arg_value = list_get_begin(arg_values);
  if (arg_value.type != TOK_LIST_END)
    T_ERROR(arg_value, "too many arguments for a macro\n");

  value = list_new();
  SUBSTITUTE({
    for (i = 0; i < args_count; i++)
      if (0==strcmp(args[i].name, token.value.t_char_p))
	{
	  temp = list_copy(args[i].value);
	  if (paste)
	    {
	      arg_value = list_get_begin(temp);
	      while (arg_value.type == TOK_SPACE)
		arg_value = list_get_begin(temp);
	      if (arg_value.type == TOK_LIST_END)
		{
		  arg_value = list_read_begin(body);
		  while (arg_value.type == TOK_SPACE)
		    arg_value = list_read_begin(body);
		}
	      paste_tokens(value, arg_name, arg_value);
	      paste = 0;
	    }
	  list_add_list_end(value, temp);
	  list_free(temp);
	  paste = 2;
	  break;
	}
    if (paste == 2)
      {
	paste = 0;
	continue;
      }
  },arg_name);
  for (i = 0; i < args_count; i++)
    {
      list_dec_refcount(args[i].value);
      list_free(args[i].value);
    }
  return value;
}

struct list_t *substitute_for_in(struct list_t *body, char *name, struct list_t *values)
{
  struct token_t token,t1,t2;
  struct list_t *value, *temp, *arg;
  int paste = 0;
  
  value= list_new();
  for (t1 = list_get_begin(values); t1.type != TOK_LIST_END; t1 = list_get_begin(values))
    {
      arg = t1.value.t_struct_list_t_p;
      
      SUBSTITUTE({
	if (0==strcmp(name, token.value.t_char_p))
	  {
	    temp = list_copy(arg);
	    if (paste)
	      {
		t2 = list_get_begin(temp);
		while (t2.type == TOK_SPACE)
		  t2 = list_get_begin(temp);
		if (t2.type == TOK_LIST_END)
		  {
		    t2 = list_read_begin(body);
		    while (t2.type == TOK_SPACE)
		      t2 = list_read_begin(body);
		  }
		paste_tokens(value, t1, t2);
		paste = 0;
	      }
	    list_add_list_end(value, temp);
	    list_free(temp);
	    continue;
	  }},t1);    
      
      list_dec_refcount(arg);
      list_free(arg);
    }
  list_dec_refcount(body);
  list_free(body);
  string_free(name);
  return value;
}

struct list_t *substitute_for_from(struct list_t *body, char *name, int from, int to, int step)
{
  struct token_t token,t1,t2;
  struct list_t *value;
  int paste = 0;
  int i;
  
  if (((to-from)*step < 0) || (step == 0))
    {
      list_dec_refcount(body);
      list_free(body);
      string_free(name);
      return list_new();
    }
  
  value = list_new();
  for (i = from; (to-i)*step >= 0; i+=step)
    {
      t2.type = TOK_INTEGER;
      t2.value.t_int = i;
      t2.position.file = "";
      t2.position.line_begin = 0;
      t2.position.line_end = 0;
      
      SUBSTITUTE({
	if (0==strcmp(name, token.value.t_char_p))
	  {
	    if (paste)
	      {
		paste_tokens(value, t1, t2);
		paste = 0;
	      }
	    else
	      list_add_end(value, t2);
	    continue;
	  }},t1);
    }
  list_dec_refcount(body);
  list_free(body);
  return value;
}

static int define_end[1] = {TOK_NEWLINE};
static int macro_end[1] = {TOK_ENDM};
static int if_end[1] = {TOK_ENDIF};
static int for_end[1] = {TOK_ENDFOR};
static int paren_end[1] = {TOK_PAREN_CLOSE};
static int brace_end[1] = {TOK_BRACE_CLOSE};
static int square_end[1] = {TOK_SQUARE_CLOSE};

#define DO_BLOCK(ENDS,ENDS_COUNT) (({if (!parser) *token = feeder_get(input); if (dest) list_add_end(dest,next); next = do_block(input,token,parser,dest,src,ENDS,ENDS_COUNT);}))
static struct token_t do_block(struct feeder_t *input, struct token_t *token, 
			       struct token_t (*parser)(struct feeder_t *input, struct token_t *token),
			       struct list_t *dest, struct list_t **src, int *ends, int ends_count)
{
  struct token_t next;
  int i;

  if (parser)
    {
      while(1)
	{
	  if (*src)
	    {
	      next = list_get_begin(*src);
	      if (next.type == TOK_LIST_END)
		{
		  list_free(*src);
		  *src = NULL;
		  continue;
		}
	      else
		break;
	    }
	  next = parser(input,token);
	  if (next.type == TOK_PC_SUCCESS)
	    continue;
	  if (next.type == TOK_LIST)
	    {
	      *src = next.value.t_struct_list_t_p;
	      continue;
	    }
	  break;
	}
    }
  else
    next = *token;

  while (1)
    {
      for (i = 0; i < ends_count; i++)
	if (next.type == ends[i])
	  return next;
      switch (next.type)
	{
	case TOK_DEFINE: DO_BLOCK(define_end,1); break;
	case TOK_MACRO: DO_BLOCK(macro_end,1); break;
	case TOK_IF: 
	case TOK_IFDEF:
	case TOK_IFNDEF: DO_BLOCK(if_end,1); break;
	case TOK_FOR: DO_BLOCK(for_end,1); break;
	case TOK_PAREN_OPEN: DO_BLOCK(paren_end,1); break;
	case TOK_BRACE_OPEN: DO_BLOCK(brace_end,1); break;
	case TOK_SQUARE_OPEN: DO_BLOCK(square_end,1); break;
	default: break;
	}
      if (dest)
	list_add_end(dest,next);
      if (parser)
	{
	  while(1)
	    {
	      if (*src)
		{
		  next = list_get_begin(*src);
		  if (next.type == TOK_LIST_END)
		    {
		      list_free(*src);
		      *src = NULL;
		      continue;
		    }
		  else
		    break;
		}
	      next = parser(input,token);
	      if (next.type == TOK_PC_SUCCESS)
		continue;
	      if (next.type == TOK_LIST)
		{
		  *src = next.value.t_struct_list_t_p;
		  continue;
		}
	      break;
	    }
	}
      else
	{
	  *token = feeder_get(input);
	  next = *token;
	}
    }
}

struct token_t parser_get_block(struct feeder_t *input, struct token_t *token, ...)
{
  int ends[10]; // 10 should be enough
  int ends_count = 0;
  struct token_t result;
  struct list_t *list;

  va_list ap;
  va_start(ap, token);
  for (ends[ends_count] = va_arg(ap,int); ends[ends_count] != TOKENS_COUNT; ends[ends_count] = va_arg(ap,int))
    {
      ends_count++;
      if (ends_count >= 10)
	ERROR("internal error\n");
    }
  va_end(ap);
  
  list = list_new();

  do_block(input,token,NULL,list,NULL,ends,ends_count);
  
  result.type = TOK_LIST;
  result.value.t_struct_list_t_p = list;
  return result;
}

struct token_t parser_skip_block(struct feeder_t *input, struct token_t *token, ...)
{
  int ends[10]; // 10 should be enough
  int ends_count = 0;
  struct token_t result;

  va_list ap;
  va_start(ap, token);
  for (ends[ends_count] = va_arg(ap,int); ends[ends_count] != TOKENS_COUNT; ends[ends_count] = va_arg(ap,int))
    {
      ends_count++;
      if (ends_count >= 10)
	ERROR("internal error\n");
    }
  va_end(ap);
  
  do_block(input,token,NULL,NULL,NULL,ends,ends_count);
  
  result.type = TOK_PC_SUCCESS;
  return result;
}

struct token_t parser_process_block(struct feeder_t *input, struct token_t *token, ...)
{
  int ends[10]; // 10 should be enough
  int ends_count = 0;
  struct token_t result;
  struct list_t *dest;
  struct list_t *src = NULL;

  va_list ap;
  va_start(ap, token);
  for (ends[ends_count] = va_arg(ap,int); ends[ends_count] != TOKENS_COUNT; ends[ends_count] = va_arg(ap,int))
    {
      ends_count++;
      if (ends_count >= 10)
	ERROR("internal error\n");
    }
  va_end(ap);
  
  dest = list_new();
  result = do_block(input,token,preprocessor_p,dest,&src,ends,ends_count);
  feeder_unget(input,*token);
  if (src)
    feeder_add_list(input,src);
  feeder_unget(input, result);
  feeder_add_list(input,dest);
  *token = feeder_get(input);
  result.type = TOK_PC_SUCCESS;
  return result;
}
