#define ZASM_PREPROCESSOR_C 1

#include <zasm/preprocessor.h>
#include <zasm/lexer.h>
#include <zasm/memory.h>
#include <zasm/string.h>
#include <zasm/parser.h>
#include <zasm/str.h>
#include <stdio.h>

struct preprocessor_t *preprocessor_new(char *path, char **include_dirs)
{
  struct preprocessor_t *val;
  ALLOCATE(struct preprocessor_t, val, 1);
  preprocessor_fill(val,path,include_dirs);
  return val;
}

void preprocessor_fill(struct preprocessor_t *pp, char *path, char **include_dirs)
{
  pp->path = path;
  pp->look_ahead.type = TOKENS_COUNT;
  pp->include_dirs = include_dirs;
  pp->ns = pp_namespace_new();
  pp_namespace_add_level(pp->ns);
  pp->expr_alloc = expr_alloc_new();
}

char *preprocessor_get_include_file(struct preprocessor_t *pp, char *name)
{
  FILE *file;
  char *filename;
  int i;
  
  for (i = 0; pp->include_dirs[i]; i++)
    {
      filename = STR_CONCAT(pp->include_dirs[i], name);
      if ((file = fopen(filename, "r")))
	if (fclose(file) == EOF)
	  ERROR("I/O error on \"%s\"\n", name);
	else
	  return filename;
      else
	free(filename);
    }
  ERROR("can not find include file \"%s\"\n", name);
  return NULL;
}

// preprocessor_execute: in pp.parser

void preprocessor_delete(struct preprocessor_t *pp)
{
  pp_namespace_del_level(pp->ns);
  if (!namespace_is_empty(&(pp->ns->namespace)))
    ERROR("reached end of input before #begin_pp_section was complete\n");
  pp_namespace_free(pp->ns);
  expr_alloc_free(pp->expr_alloc);
}

void preprocessor_free(struct preprocessor_t *pp)
{
  preprocessor_delete(pp);
  free(pp);
}
