#define ZASM_TOKENS_C 1

#include <zasm/tokens.h>
#include <zasm/string.h>
#include <zasm/errors.h>

struct token_t token_copy(struct token_t token)
{
  switch (token.type)
    {
      // will be auto-gen
case TOK_STRING:
case TOK_ISTRING:
case TOK_NAME:
case TOK_VARARG:
case TOK_END_FILE:

      token.value.t_char_p = string_copy(token.value.t_char_p);
      break;
    default:
      break;
    }
  return token;
}

void token_print(struct token_t token, FILE *file)
{
#define WRITE(FMT,A...) fprintf(file, FMT, ## A)
  switch (token.type)
    {
case TOK_MULT_LIST: break;
case TOK_PROCESSOR: break;
case TOK_PROC_INSTR: break;
case TOK_PROC_INSTR_ARG: break;
case TOK_LITTLE_ENDIAN: {WRITE("LITTLE ENDIAN");}break;
case TOK_BIG_ENDIAN: {WRITE("BIG ENDIAN");}break;
case TOK_BYTE_INT: {WRITE("`%02hhX`", token.value.t_int);}break;
case TOK_WORD_INT: {WRITE("`%04hX`", token.value.t_int);}break;
case TOK_DWORD_INT: {WRITE("`%08lX`", token.value.t_int);}break;
case TOK_BYTE_INT_P: {WRITE("`%02hhX`", *(token.value.t_int_p));}break;
case TOK_WORD_INT_P: {WRITE("`%04hX`", *(token.value.t_int_p));}break;
case TOK_DWORD_INT_P: {WRITE("`%08lX`", *(token.value.t_int_p));}break;
case TOK_BYTE: {WRITE("#byte ");}break;
case TOK_WORD: {WRITE("#word ");}break;
case TOK_DWORD: {WRITE("#dword ");}break;
case TOK_ARG: {WRITE("@%i", token.value.t_int);}break;
case TOK_DB: {WRITE(".db ");}break;
case TOK_DW: {WRITE(".dw ");}break;
case TOK_DD: {WRITE(".dd ");}break;
case TOK_NONE: break;
case TOK_LIST_END: break;
case TOK_STREAM_END: break;
case TOK_END_OF_FILE: break;
case TOK_STRING: {WRITE("\"%s\"", token.value.t_char_p);}break;
case TOK_ISTRING: {WRITE("<%s>", token.value.t_char_p);}break;
case TOK_NAME: {WRITE("%s", token.value.t_char_p);}break;
case TOK_VARARG: {WRITE("%s...", token.value.t_char_p);}break;
case TOK_NEWLINE: {WRITE("\n");}break;
case TOK_ESCAPED_NEWLINE: {WRITE("\n");}break;
case TOK_SPACE: {WRITE(" ");}break;
case TOK_SPECIAL: {WRITE("%c", token.value.t_char);}break;
case TOK_INTEGER: {WRITE("%i", token.value.t_int);}break;
case TOK_ALIGN_ADDR: {WRITE("#align_addr ");}break;
case TOK_ALIGN_MULT: {WRITE("#align_mult ");}break;
case TOK_ORG: {WRITE("#org ");}break;
case TOK_PAGE: {WRITE("#setpage ");}break;
case TOK_PAGE_START: {WRITE("#page_start ");}break;
case TOK_NEWPAGE: {WRITE("#newpage ");}break;
case TOK_EQU: {WRITE("#equ ");}break;
case TOK_EXPORT: {WRITE("#export ");}break;
case TOK_BEGIN_SECTION: {WRITE("#begin_section ");}break;
case TOK_END_SECTION: {WRITE("#end_section ");}break;
case TOK_SET_PAGE: break;
case TOK_PC_FAIL: {WRITE("<pc fail>");}break;
case TOK_PC_FAIL1: {WRITE("<pc fail1>");}break;
case TOK_PC_SUCCESS: {WRITE("<pc success>");}break;
case TOK_BACKSLASH: {WRITE("\\");}break;
case TOK_EXPR: {WRITE("<expr>");}break;
case TOK_LIST: {WRITE("<list>");}break;
case TOK_END_FILE: {WRITE("<endFile>");}break;
case TOK_OP_IFTHEN: {WRITE("?");}break;
case TOK_OP_ELSE: {WRITE(":");}break;
case TOK_OP_LOG_OR: {WRITE("||");}break;
case TOK_OP_LOG_AND: {WRITE("&&");}break;
case TOK_OP_LOG_XOR: {WRITE("^^");}break;
case TOK_OP_LOG_NOT: {WRITE("!");}break;
case TOK_OP_BIN_OR: {WRITE(".|");}break;
case TOK_OP_BIN_AND: {WRITE(".&");}break;
case TOK_OP_BIN_XOR: {WRITE(".^");}break;
case TOK_OP_BIN_NOT: {WRITE(".~");}break;
case TOK_OP_SHIFT_LEFT: {WRITE("<<");}break;
case TOK_OP_SHIFT_RIGHT: {WRITE(">>");}break;
case TOK_OP_EQU: {WRITE("==");}break;
case TOK_OP_NEQ: {WRITE("!=");}break;
case TOK_OP_LT: {WRITE("<");}break;
case TOK_OP_GT: {WRITE(">");}break;
case TOK_OP_LE: {WRITE("<=");}break;
case TOK_OP_GE: {WRITE(">=");}break;
case TOK_OP_CMP: {WRITE("<=>");}break;
case TOK_OP_ADD: {WRITE("+");}break;
case TOK_OP_SUB: {WRITE("-");}break;
case TOK_OP_MUL: {WRITE("*");}break;
case TOK_OP_DIV: {WRITE("/");}break;
case TOK_OP_MOD: {WRITE("%%");}break;
case TOK_OP_POW: {WRITE("**");}break;
case TOK_DOT: {WRITE(".");}break;
case TOK_PAREN_OPEN: {WRITE("(");}break;
case TOK_PAREN_CLOSE: {WRITE(")");}break;
case TOK_BRACE_OPEN: {WRITE("{");}break;
case TOK_BRACE_CLOSE: {WRITE("}");}break;
case TOK_SQUARE_OPEN: {WRITE("[");}break;
case TOK_SQUARE_CLOSE: {WRITE("]");}break;
case TOK_COMMA: {WRITE(", ");}break;
case TOK_PASTE: {WRITE(" ## ");}break;
case TOK_DEFINED: {WRITE("defined");}break;
case TOK_DEFINE: {WRITE("#define");}break;
case TOK_MACRO: {WRITE(".macro");}break;
case TOK_REDEF: {WRITE("#redef");}break;
case TOK_UNDEF: {WRITE("#undef");}break;
case TOK_ENDM: {WRITE("#endm");}break;
case TOK_PP_BEGIN_SECTION: {WRITE("#pp_begin_section");}break;
case TOK_PP_END_SECTION: {WRITE("#pp_end_section");}break;
case TOK_INCLUDE: {WRITE("#include");}break;
case TOK_IF: {WRITE("#if");}break;
case TOK_THEN: {WRITE("#then");}break;
case TOK_IFDEF: {WRITE("#ifdef");}break;
case TOK_IFNDEF: {WRITE("#ifndef");}break;
case TOK_ELSE: {WRITE("#else");}break;
case TOK_ELSEIF: {WRITE("#elseif");}break;
case TOK_ELSEIFDEF: {WRITE("#elseifdef");}break;
case TOK_ELSEIFNDEF: {WRITE("#elseifndef");}break;
case TOK_ENDIF: {WRITE("#endif");}break;
case TOK_FOR: {WRITE("#for");}break;
case TOK_IN: {WRITE("#in");}break;
case TOK_FROM: {WRITE("#from");}break;
case TOK_TO: {WRITE("#to");}break;
case TOK_STEP: {WRITE("#step");}break;
case TOK_DO: {WRITE("#do");}break;
case TOK_ENDFOR: {WRITE("#endfor");}break;
case TOK_ERROR: {WRITE("#error");}break;
case TOK_WARNING: {WRITE("#warning");}break;
case TOK_DEBUG: {WRITE("#debug");}break;
case TOK_FEED_ACTION: break;
case TOK_FEED_LIST: break;
case TOK_OUTPUT: {WRITE("#output ");}break;

    default:
      ERROR("unknown token\n");
      break;
    }
  return;
}

void token_free(struct token_t token)
{
  switch (token.type)
    {
      //will be auto-gen
case TOK_STRING:
case TOK_ISTRING:
case TOK_NAME:
case TOK_VARARG:
case TOK_END_FILE:

      string_free(token.value.t_char_p);
      break;
    default:
      break;
    }
}
