#define ZASM_C 1

#include <zasm/errors.h>
#include "zasm.h"
#include <zasm/list.h>
#include <zasm/stream.h>
#include <zasm/memory.h>
#include <zasm/lexer.h>
#include <zasm/preprocessor.h>
#include <zasm/assembler.h>
#include <zasm/linker.h>
#include <zasm/writer.h>
#include <zasm/intel_hex_writer.h>
#include <zasm/str.h>
#include <string.h>

//#define FD DEBUG("\"%s\".%i: %s\n", __FILE__, __LINE__, __FUNCTION__)
//#define FD printf("%s.%i:%s\n", __FILE__, __LINE__, __FUNCTION__);
#define FD

//#define ZASM_PREFIX "/Bestanden/Bestanden/Programmeren/zasm-0.1"

#define IS_OPTION_ARG(SHORT, LONG) ((arg=arg_cmp_val_noequ(argc,argv,&i,SHORT))||(arg=arg_cmp_val(argc,argv,&i,LONG)))
#define IS_OPTION(SHORT, LONG) ((arg_cmp(argc,argv,&i,SHORT))||(arg_cmp(argc,argv,&i,LONG)))

#define FILE_ADD(TYPE,NAME) (({if (files_count >= files_size)\
	    {\
	      files_size+=10;\
	      REALLOCATE(char *, file_names, files_size);\
	      REALLOCATE(int, file_types, files_size);\
	    }\
	  file_types[files_count] = TYPE;\
	  file_names[files_count] = NAME;\
	  files_count++;\
}))

#define INCLUDE_ADD(NAME) (({if (includes_count >= includes_size)\
	    {\
	      includes_size+=10;\
	      REALLOCATE(char *, includes, includes_size);\
	    }\
	  includes[includes_count] = NAME;\
	  includes_count++;\
}))

static char *change_ext(int flags, char *base)
{
  char *dest;
  char *ext;
  int i, l;

  FD;

  l = 0;
  for (i = 0; base[i]; i++)
    if (base[i] == '.')
      l = i;
  if (!l)
    l = i;
  
  if (flags&ZASM_LINKER)
    ext = ".hex";
  else if (flags&ZASM_ASSEMBLER)
    ext = ".o";
  else if (flags&ZASM_PREPROCESSOR)
    ext = ".pp";
  else
    ERROR("nothing to do\n");
  
  if (0==strcmp(ext, base+l))
    ext = ".out";

  ALLOCATE(char, dest, l+strlen(ext)+1);
  strncpy(dest, base, l);
  strcpy(dest+l, ext);
  return dest;
}

static inline char *arg_cmp_val(int argc, char **argv, int *i, char *name)
{
  char *arg;
  
  FD;

  for (arg = argv[*i]; 1; arg++,name++)
    {
      if (*name == 0)
	if (*arg == '=')
	  return arg+1;
	else if (*arg == 0)
	  {
	    (*i)++;
	    if (*i < argc)
	      return argv[*i];
	    else
	      ERROR("expecting a value for option \"%s\"\n", name);
	  }
	else
	  return NULL;
      else if (*arg != *name)
	return NULL;
    }  
}

static inline char *arg_cmp_val_noequ(int argc, char **argv, int *i, char *name)
{
  char *arg;
  
  FD;

  for (arg = argv[*i]; 1; arg++,name++)
    {
      if (*name == 0)
	if (*arg == '=')
	  return arg+1;
	else if (*arg == 0)
	  {
	    (*i)++;
	    if (*i < argc)
	      return argv[*i];
	    else
	      ERROR("expecting a value for option \"%s\"\n", name);
	  }
	else
	  return arg;
      else if (*arg != *name)
	return NULL;
    }  
}

static inline int arg_cmp(int argc, char **argv, int *i, char *name)
{
  char *arg;
  
  FD;

  for (arg = argv[*i]; 1; arg++,name++)
    {
      if (*name == 0)
	if (*arg == 0)
	  return 1;
	else
	  return 0;
      else if (*arg != *name)
	return 0;
    }  
}

void list_add_def_directive(int type, struct list_t *list, char *definition)
{
  struct feeder_t *input;
  struct string_lexer_t *lex;
  struct token_t token;
  int equ_passed = 0;

  FD;

  lex = string_lexer_new(definition);
  input = feeder_new();
  feeder_add_action(input, action_new(ACTION_EXECUTE_T(string_lexer_execute), NULL, NULL, lex));
  token.type = type;
  token.position.file = string_new("no file");
  token.position.line_begin = 0;
  token.position.line_end = 0;
  for (; token.type != TOK_END_OF_FILE; token = feeder_get(input))
    {
      if ((!equ_passed) && (token.type == TOK_OP_EQU))
	{
	  token.type = TOK_SPACE;
	  equ_passed = 1;
	}
      list_add_end(list, token);
    }
  token.type = TOK_NEWLINE;
  list_add_end(list, token);
}

void list_add_include_directive(int type, struct list_t *list, char *definition)
{
  struct feeder_t *input;
  struct string_lexer_t *lex;
  struct token_t token;

  FD;

  lex = string_lexer_new(definition);
  input = feeder_new();
  feeder_add_action(input, action_new(ACTION_EXECUTE_T(string_lexer_execute), NULL, NULL, lex));
  token.type = type;
  token.position.file = string_new("no file");
  token.position.line_begin = 0;
  token.position.line_end = 0;
  token.type = TOK_STRING;
  token.value.t_char_p = definition;
  list_add_end(list, token);
  token.type = TOK_NEWLINE;
  list_add_end(list, token);
}

static struct action_t *begin_lexer(char **file_names, int *file_type, int file_count)
{
  FD;

  if ((file_count != 1) || (file_type[0] != FILE_INPUT))
    ERROR("nothing to do\n");
  return action_new(ACTION_EXECUTE_T(lexer_execute), NULL, NULL, lexer_new(file_names[0]));
}

static struct action_t *begin_preprocessor(int flags, struct list_t *init, char **file_names, int *file_types, int file_count,
				    char **include_paths)
{
  int i;
  char *first = NULL;
  struct token_t t;
  struct list_t *list;
  struct feeder_t *feeder;
  
  FD;

  if (!(flags&ZASM_PREPROCESSOR))
    return begin_lexer(file_names, file_types, file_count);
  list = list_copy(init);
  t.position.file = string_new("no file");
  t.position.line_begin = 0;
  t.position.line_end = 0;
  for (i = 0; i < file_count; i++)
    {
      if (file_types[i] == FILE_INPUT)
	{
	  if (!first)
	    {
	      t.type = TOK_FEED_ACTION;
	      t.value.t_struct_action_t_p = begin_lexer(&(file_names[i]), &(file_types[i]), 1);
	      list_add_end(list, t);
	      first = file_names[i];
	    }
	  else
	    {
	      t.type = TOK_INCLUDE;
	      list_add_end(list, t);
	      t.type = TOK_STRING;
	      t.value.t_char_p = file_names[i];
	      list_add_end(list, t);
	      t.type = TOK_NEWLINE;
	      list_add_end(list, t);
	    }
	}
      else if (file_types[i] == FILE_OUTPUT)
	{
	  t.type = TOK_OUTPUT;
	  list_add_end(list, t);
	  t.type = TOK_STRING;
	  t.value.t_char_p = file_names[i];
	  list_add_end(list, t);
	  t.type = TOK_NEWLINE;
	  list_add_end(list, t);
	}
    }
  feeder = feeder_new();
  feeder_add_list(feeder, list);
  return action_new(ACTION_EXECUTE_T(preprocessor_execute), feeder, NULL, 
		    preprocessor_new(str_get_path(first), include_paths));
}

static struct action_t *begin_assembler(int flags, struct list_t *init, char **file_names, int *file_types, int file_count,
				 struct processor_t *proc,
				 struct list_t *pp_init, char **include_paths)
{
  int i;
  struct token_t t;
  struct list_t *list;
  struct feeder_t *feeder;
  
  FD;

  if (!(flags&ZASM_ASSEMBLER))
    return begin_preprocessor(flags, pp_init, file_names, file_types, file_count, include_paths);
  list = list_copy(init);
  t.position.file = string_new("no file");
  t.position.line_begin = 0;
  t.position.line_end = 0;
  for (i = 0; i < file_count; i++)
    {
      if (file_types[i] == FILE_INPUT)
	{
	  t.type = TOK_FEED_ACTION;
	  t.value.t_struct_action_t_p = begin_preprocessor(flags, pp_init, &(file_names[i]), &(file_types[i]), 1,
							   include_paths);
	  list_add_end(list, t);
	}
      else if (file_types[i] == FILE_OUTPUT)
	{
	  t.type = TOK_OUTPUT;
	  list_add_end(list, t);
	  t.type = TOK_STRING;
	  t.value.t_char_p = file_names[i];
	  list_add_end(list, t);
	  t.type = TOK_NEWLINE;
	  list_add_end(list, t);
	}
    }
  feeder = feeder_new();
  feeder_add_list(feeder, list);
  return action_new(ACTION_EXECUTE_T(assembler_execute), feeder, NULL, 
		    assembler_new(proc));
}

static struct action_t *begin_linker(int flags, struct list_t *init, char **file_names, int *file_types, int file_count,
				     char *dump_file, char *export_file,
				     struct list_t *asm_init, struct processor_t *proc,
				     struct list_t *pp_init, char **include_paths
				     )
{
  int i;
  struct token_t t;
  struct list_t *list;
  struct feeder_t *feeder;
  
  FD;

  if (!(flags&ZASM_LINKER))
    return begin_assembler(flags, asm_init, file_names, file_types, file_count, proc, pp_init, include_paths);
  list = list_copy(init);
  t.position.file = string_new("no file");
  t.position.line_begin = 0;
  t.position.line_end = 0;
  for (i = 0; i < file_count; i++)
    {
      if (file_types[i] == FILE_INPUT)
	{
	  t.type = TOK_BEGIN_SECTION;
	  list_add_end(list, t);
	  t.type = TOK_FEED_ACTION;
	  t.value.t_struct_action_t_p = begin_assembler(flags, asm_init, &(file_names[i]), &(file_types[i]), 1,
							proc, pp_init, include_paths);
	  list_add_end(list, t);
	  t.type = TOK_END_SECTION;
	  list_add_end(list, t);
	}
      else if (file_types[i] == FILE_OUTPUT)
	{
	  t.type = TOK_OUTPUT;
	  list_add_end(list, t);
	  t.type = TOK_STRING;
	  t.value.t_char_p = file_names[i];
	  list_add_end(list, t);
	  t.type = TOK_NEWLINE;
	  list_add_end(list, t);
	}
    }
  feeder = feeder_new();
  feeder_add_list(feeder, list);
  return action_new(ACTION_EXECUTE_T(linker_execute), feeder, NULL, 
		    linker_new(dump_file, export_file));
}

int main(int argc, char *argv[])
{
  int flags = ZASM_PREPROCESSOR | ZASM_ASSEMBLER | ZASM_LINKER;

  int includes_count = 0;
  int includes_size = 0;
  char **includes = NULL;

  int files_count = 0;
  int files_size = 0;
  char **file_names = NULL;
  int *file_types = NULL;

  char *dump_file = NULL;
  char *export_file = NULL;

  struct list_t *pp_init = list_new();
  struct list_t *asm_init = list_new();
  struct list_t *lnk_init = list_new();

  char *processor = DATADIR "/ZASM80.TAB";

  int i, n, j;
  char *arg;
  struct processor_t *proc = NULL;
  struct writer_t *writer = NULL;
  struct intel_hex_writer_t *ihw = NULL;
  struct feeder_t *feeder;

  FD;

  error_init(&argc, &argv);
  INCLUDE_ADD(DATADIR);
  for (i = 1; i < argc; i++)
    {
      if (IS_OPTION_ARG("-o","--output"))
	FILE_ADD(FILE_OUTPUT, arg);
      else if (IS_OPTION_ARG("-D","--define"))
	list_add_def_directive(TOK_DEFINE, pp_init, arg);
      else if (IS_OPTION_ARG("-e","--equ"))
	list_add_def_directive(TOK_EQU, lnk_init, arg);
      else if (IS_OPTION_ARG("-I", "--include_path"))
	INCLUDE_ADD(arg);
      else if (IS_OPTION_ARG("-i", "--include"))
	list_add_include_directive(TOK_INCLUDE, lnk_init, arg);
      else if (IS_OPTION_ARG("-p", "--processor"))
	processor = arg;
      else if (IS_OPTION_ARG("-x", "--export"))
	export_file = arg;
      else if (IS_OPTION_ARG("-d", "--dump"))
	dump_file = arg;
      else if (IS_OPTION("-E", "--preprocessor"))
	flags = ZASM_PREPROCESSOR;
      else if (IS_OPTION("-a", "--assembler"))
	flags = ZASM_PREPROCESSOR | ZASM_ASSEMBLER;
      else if (IS_OPTION("-np","--no-preprocessor"))
	flags = flags&(~ZASM_PREPROCESSOR);
      else if (IS_OPTION("-ap","--add-preprocessor"))
	flags |= ZASM_PREPROCESSOR;
      else if (IS_OPTION("-na","--no-assembler"))
	flags = flags&(~ZASM_ASSEMBLER);
      else if (IS_OPTION("-aa","--add-assembler"))
	flags |= ZASM_ASSEMBLER;
      else if (IS_OPTION("-nl","--no-linker"))
	flags = flags&(~ZASM_LINKER);
      else if (IS_OPTION("-al","--add-linker"))
	flags |= ZASM_LINKER;
      else if (IS_OPTION("-h","--help"))
	{
	  fprintf(stderr,"usage: zasm option ...\n");
	  fprintf(stderr,"option is one of:\n");
	  fprintf(stderr," (-o | --output) <file>         : output to <file>\n");
	  fprintf(stderr," (-x | --export) <file>         : write an include file with all exported labels to <file>\n");
	  fprintf(stderr," (-d | --dump) <file>           : dump all defined labels to <file>, for use with a debugger\n");
	  fprintf(stderr," (-D | --define) <name>[=<def>] : define the macro <name> to be <def>\n");
	  fprintf(stderr," (-e | --equ) <name>[=<def>]    : define the label <name> to be <def>\n");
	  fprintf(stderr," (-I | --include-path) <path>   : add <path> to the include path\n");
	  fprintf(stderr," (-i | --include) <file>        : include <file>\n");
	  fprintf(stderr," (-E | --preprocessor) <file>   : execute only the preprocessor\n");
	  fprintf(stderr," (-a | --assembler) <file>      : execute the preprocessor and the assembler, but not the linker\n");
	  fprintf(stderr," <file>                         : add <file> to the input\n");
	  exit(0);
	}
      else if (IS_OPTION("-v","--version"))
	{
	  fprintf(stderr, "zasm (ZASM) 0.1 alpha 0");
	  exit(0);
	}
      else
	FILE_ADD(FILE_INPUT, argv[i]);
    }
  INCLUDE_ADD(NULL);

  if (files_count == 0)
    ERROR("no input files\n");
  
  n = 0;
  for (i = 0; i < files_count; i++)
    if (file_types[i] == FILE_OUTPUT)
      {
	n++;
	j = i;
      }
  if (files_count ==  n)
    ERROR("no input files\n");
  if (n == 0)
    arg = change_ext(flags, file_names[0]);
  else if (n == 1)
    {
      arg = file_names[j];
      file_types[j] = FILE_IGNORE;
    }
  else
    {
      if (flags&ZASM_LINKER)
	arg = NULL;
      else
	ERROR("multiple outputs given for an action that accepts only one\n");
    }
  
  proc = processor_read(processor);
  feeder = feeder_new();
  feeder_add_action(feeder, begin_linker(flags, lnk_init, file_names, file_types, files_count,
					 dump_file, export_file,
					 asm_init, proc,
					 pp_init, includes));
  FD;
  if (flags&ZASM_LINKER)
    {
      ihw = intel_hex_writer_new(arg, proc->endianness);
      FD;
      intel_hex_writer_execute(feeder, NULL, ihw);
      FD;
    }
  else
    {
      writer = writer_new(arg);
      FD;
      writer_execute(feeder, NULL, writer);
      FD;
    }
  return 0;
}
